@@ -2,6 +2,7 @@
 obtained from config-files, forms, external services or command-line
 parsing, converted from JSON/YAML (or something else) to Python data-types."""
 
+import inspect
 import re
 
 try:
@@ -9,7 +10,8 @@ try:
 except ImportError:
     from contextlib2 import ExitStack
 
-__version__ = "0.7.2"
+
+__version__ = "0.7.5"
 __all__ = [
     "Schema",
     "And",
@@ -117,7 +119,7 @@ class And(object):
         """The provided parameters"""
         return self._args
 
-    def validate(self, data):
+    def validate(self, data, **kwargs):
         """
         Validate data using defined sub schema/expressions ensuring all
         values are valid.
@@ -125,7 +127,7 @@ class And(object):
         :return: returns validated data
         """
         for s in [self._schema(s, error=self._error, ignore_extra_keys=self._ignore_extra_keys) for s in self._args]:
-            data = s.validate(data)
+            data = s.validate(data, **kwargs)
         return data
 
 
@@ -144,7 +146,7 @@ class Or(And):
         if failed:
             raise SchemaOnlyOneAllowedError(["There are multiple keys present " + "from the %r condition" % self])
 
-    def validate(self, data):
+    def validate(self, data, **kwargs):
         """
         Validate data using sub defined schema/expressions ensuring at least
         one value is valid.
@@ -154,13 +156,14 @@ class Or(And):
         autos, errors = [], []
         for s in [self._schema(s, error=self._error, ignore_extra_keys=self._ignore_extra_keys) for s in self._args]:
             try:
-                validation = s.validate(data)
+                validation = s.validate(data, **kwargs)
                 self.match_count += 1
                 if self.match_count > 1 and self.only_one:
                     break
                 return validation
             except SchemaError as _x:
-                autos, errors = _x.autos, _x.errors
+                autos += _x.autos
+                errors += _x.errors
         raise SchemaError(
             ["%r did not validate %r" % (self, data)] + autos,
             [self._error.format(data) if self._error else None] + errors,
@@ -187,7 +190,9 @@ class Regex(object):
 
     def __init__(self, pattern_str, flags=0, error=None):
         self._pattern_str = pattern_str
-        flags_list = [Regex.NAMES[i] for i, f in enumerate("{0:09b}".format(flags)) if f != "0"]  # Name for each bit
+        flags_list = [
+            Regex.NAMES[i] for i, f in enumerate("{0:09b}".format(int(flags))) if f != "0"
+        ]  # Name for each bit
 
         if flags_list:
             self._flags_names = ", flags=" + "|".join(flags_list)
@@ -205,7 +210,7 @@ class Regex(object):
         """The pattern for the represented regular expression"""
         return self._pattern_str
 
-    def validate(self, data):
+    def validate(self, data, **kwargs):
         """
         Validated data using defined regex.
         :param data: data to be validated
@@ -217,7 +222,7 @@ class Regex(object):
             if self._pattern.search(data):
                 return data
             else:
-                raise SchemaError("%r does not match %r" % (self, data), e)
+                raise SchemaError("%r does not match %r" % (self, data), e.format(data) if e else None)
         except TypeError:
             raise SchemaError("%r is not string nor buffer" % data, e)
 
@@ -237,7 +242,7 @@ class Use(object):
     def __repr__(self):
         return "%s(%r)" % (self.__class__.__name__, self._callable)
 
-    def validate(self, data):
+    def validate(self, data, **kwargs):
         try:
             return self._callable(data)
         except SchemaError as x:
@@ -268,6 +273,13 @@ def _priority(s):
         return COMPARABLE
 
 
+def _invoke_with_optional_kwargs(f, **kwargs):
+    s = inspect.signature(f)
+    if len(s.parameters) == 0:
+        return f()
+    return f(**kwargs)
+
+
 class Schema(object):
     """
     Entry point of the library, use this class to instantiate validation
@@ -318,12 +330,12 @@ class Schema(object):
         """Return True if the given key is optional (does not have to be found)"""
         return any(isinstance(s, optional_type) for optional_type in [Optional, Hook])
 
-    def is_valid(self, data):
+    def is_valid(self, data, **kwargs):
         """Return whether the given data has passed all the validations
         that were specified in the given schema.
         """
         try:
-            self.validate(data)
+            self.validate(data, **kwargs)
         except SchemaError:
             return False
         else:
@@ -338,7 +350,7 @@ class Schema(object):
             message = "{0!r} {1!s}".format(self._name, message)
         return message
 
-    def validate(self, data):
+    def validate(self, data, **kwargs):
         Schema = self.__class__
         s = self._schema
         e = self._error
@@ -349,12 +361,12 @@ class Schema(object):
 
         flavor = _priority(s)
         if flavor == ITERABLE:
-            data = Schema(type(s), error=e).validate(data)
+            data = Schema(type(s), error=e).validate(data, **kwargs)
             o = Or(*s, error=e, schema=Schema, ignore_extra_keys=i)
-            return type(data)(o.validate(d) for d in data)
+            return type(data)(o.validate(d, **kwargs) for d in data)
         if flavor == DICT:
             exitstack = ExitStack()
-            data = Schema(dict, error=e).validate(data)
+            data = Schema(dict, error=e).validate(data, **kwargs)
             new = type(data)()  # new - is a dict of the validated values
             coverage = set()  # matched schema keys
             # for each key and value find a schema entry matching them, if any
@@ -370,7 +382,7 @@ class Schema(object):
                     for skey in sorted_skeys:
                         svalue = s[skey]
                         try:
-                            nkey = Schema(skey, error=e).validate(key)
+                            nkey = Schema(skey, error=e).validate(key, **kwargs)
                         except SchemaError:
                             pass
                         else:
@@ -384,17 +396,17 @@ class Schema(object):
                                 # value has a certain type, and allowing Forbidden to
                                 # work well in combination with Optional.
                                 try:
-                                    nvalue = Schema(svalue, error=e).validate(value)
+                                    nvalue = Schema(svalue, error=e).validate(value, **kwargs)
                                 except SchemaError:
                                     continue
                                 skey.handler(nkey, data, e)
                             else:
                                 try:
-                                    nvalue = Schema(svalue, error=e, ignore_extra_keys=i).validate(value)
+                                    nvalue = Schema(svalue, error=e, ignore_extra_keys=i).validate(value, **kwargs)
                                 except SchemaError as x:
                                     k = "Key '%s' error:" % nkey
                                     message = self._prepend_schema_name(k)
-                                    raise SchemaError([message] + x.autos, [e] + x.errors)
+                                    raise SchemaError([message] + x.autos, [e.format(data) if e else None] + x.errors)
                                 else:
                                     new[nkey] = nvalue
                                     coverage.add(skey)
@@ -405,7 +417,7 @@ class Schema(object):
                 s_missing_keys = ", ".join(repr(k) for k in sorted(missing_keys, key=repr))
                 message = "Missing key%s: %s" % (_plural_s(missing_keys), s_missing_keys)
                 message = self._prepend_schema_name(message)
-                raise SchemaMissingKeyError(message, e)
+                raise SchemaMissingKeyError(message, e.format(data) if e else None)
             if not self._ignore_extra_keys and (len(new) != len(data)):
                 wrong_keys = set(data.keys()) - set(new.keys())
                 s_wrong_keys = ", ".join(repr(k) for k in sorted(wrong_keys, key=repr))
@@ -414,9 +426,9 @@ class Schema(object):
                 raise SchemaWrongKeyError(message, e.format(data) if e else None)
 
             # Apply default-having optionals that haven't been used:
-            defaults = set(k for k in s if type(k) is Optional and hasattr(k, "default")) - coverage
+            defaults = set(k for k in s if isinstance(k, Optional) and hasattr(k, "default")) - coverage
             for default in defaults:
-                new[default.key] = default.default() if callable(default.default) else default.default
+                new[default.key] = _invoke_with_optional_kwargs(default.default, **kwargs) if callable(default.default) else default.default
 
             return new
         if flavor == TYPE:
@@ -428,27 +440,27 @@ class Schema(object):
                 raise SchemaUnexpectedTypeError(message, e.format(data) if e else None)
         if flavor == VALIDATOR:
             try:
-                return s.validate(data)
+                return s.validate(data, **kwargs)
             except SchemaError as x:
-                raise SchemaError([None] + x.autos, [e] + x.errors)
+                raise SchemaError([None] + x.autos, [e.format(data) if e else None] + x.errors)
             except BaseException as x:
                 message = "%r.validate(%r) raised %r" % (s, data, x)
                 message = self._prepend_schema_name(message)
-                raise SchemaError(message, self._error.format(data) if self._error else None)
+                raise SchemaError(message, e.format(data) if e else None)
         if flavor == CALLABLE:
             f = _callable_str(s)
             try:
                 if s(data):
                     return data
             except SchemaError as x:
-                raise SchemaError([None] + x.autos, [e] + x.errors)
+                raise SchemaError([None] + x.autos, [e.format(data) if e else None] + x.errors)
             except BaseException as x:
                 message = "%s(%r) raised %r" % (f, data, x)
                 message = self._prepend_schema_name(message)
-                raise SchemaError(message, self._error.format(data) if self._error else None)
+                raise SchemaError(message, e.format(data) if e else None)
             message = "%s(%r) should evaluate to True" % (f, data)
             message = self._prepend_schema_name(message)
-            raise SchemaError(message, e)
+            raise SchemaError(message, e.format(data) if e else None)
         if s == data:
             return data
         else:
@@ -456,9 +468,8 @@ class Schema(object):
             message = self._prepend_schema_name(message)
             raise SchemaError(message, e.format(data) if e else None)
 
-    def json_schema(self, schema_id, use_refs=False):
+    def json_schema(self, schema_id, use_refs=False, **kwargs):
         """Generate a draft-07 JSON schema dict representing the Schema.
-        This method can only be called when the Schema's value is a dict.
         This method must be called with a schema_id.
 
         :param schema_id: The value of the $id on the main schema
@@ -532,90 +543,91 @@ class Schema(object):
 
             return_schema = {}
 
-            is_a_ref = allow_reference and schema.as_reference
-            if schema.description and not is_a_ref:
-                return_schema["description"] = schema.description
-            if description and not is_a_ref:
-                return_schema["description"] = description
+            return_description = description or schema.description
+            if return_description:
+                return_schema["description"] = return_description
 
-            if flavor != DICT and is_main_schema:
-                raise ValueError("The main schema must be a dict.")
+            # Check if we have to create a common definition and use as reference
+            if allow_reference and schema.as_reference:
+                # Generate sub schema if not already done
+                if schema.name not in definitions_by_name:
+                    definitions_by_name[schema.name] = {}  # Avoid infinite loop
+                    definitions_by_name[schema.name] = _json_schema(schema, is_main_schema=False, allow_reference=False)
 
-            if flavor == TYPE:
-                # Handle type
-                return_schema["type"] = _get_type_name(s)
-            elif flavor == ITERABLE:
-                # Handle arrays or dict schema
-
-                return_schema["type"] = "array"
-                if len(s) == 1:
-                    return_schema["items"] = _json_schema(_to_schema(s[0], i), is_main_schema=False)
-                elif len(s) > 1:
-                    return_schema["items"] = _json_schema(Schema(Or(*s)), is_main_schema=False)
-            elif isinstance(s, Or):
-                # Handle Or values
-
-                # Check if we can use an enum
-                if all(priority == COMPARABLE for priority in [_priority(value) for value in s.args]):
-                    or_values = [str(s) if isinstance(s, Literal) else s for s in s.args]
-                    # All values are simple, can use enum or const
-                    if len(or_values) == 1:
-                        return_schema["const"] = _to_json_type(or_values[0])
-                        return return_schema
-                    return_schema["enum"] = or_values
-                else:
-                    # No enum, let's go with recursive calls
-                    any_of_values = []
-                    for or_key in s.args:
-                        new_value = _json_schema(_to_schema(or_key, i), is_main_schema=False)
-                        if new_value != {} and new_value not in any_of_values:
-                            any_of_values.append(new_value)
-                    if len(any_of_values) == 1:
-                        # Only one representable condition remains, do not put under oneOf
-                        return_schema.update(any_of_values[0])
-                    else:
-                        return_schema["anyOf"] = any_of_values
-            elif isinstance(s, And):
-                # Handle And values
-                all_of_values = []
-                for and_key in s.args:
-                    new_value = _json_schema(_to_schema(and_key, i), is_main_schema=False)
-                    if new_value != {} and new_value not in all_of_values:
-                        all_of_values.append(new_value)
-                if len(all_of_values) == 1:
-                    # Only one representable condition remains, do not put under allOf
-                    return_schema.update(all_of_values[0])
-                else:
-                    return_schema["allOf"] = all_of_values
-            elif flavor == COMPARABLE:
-                return_schema["const"] = _to_json_type(s)
-            elif flavor == VALIDATOR and type(s) == Regex:
-                return_schema["type"] = "string"
-                return_schema["pattern"] = s.pattern_str
+                return_schema["$ref"] = "#/definitions/" + schema.name
             else:
-                if flavor != DICT:
-                    # If not handled, do not check
-                    return return_schema
+                if flavor == TYPE:
+                    # Handle type
+                    return_schema["type"] = _get_type_name(s)
+                elif flavor == ITERABLE:
+                    # Handle arrays or dict schema
 
-                # Schema is a dict
+                    return_schema["type"] = "array"
+                    if len(s) == 1:
+                        return_schema["items"] = _json_schema(_to_schema(s[0], i), is_main_schema=False)
+                    elif len(s) > 1:
+                        return_schema["items"] = _json_schema(Schema(Or(*s)), is_main_schema=False)
+                elif isinstance(s, Or):
+                    # Handle Or values
 
-                # Check if we have to create a common definition and use as reference
-                if is_a_ref:
-                    # Generate sub schema if not already done
-                    if schema.name not in definitions_by_name:
-                        definitions_by_name[schema.name] = {}  # Avoid infinite loop
-                        definitions_by_name[schema.name] = _json_schema(
-                            schema, is_main_schema=False, allow_reference=False
-                        )
-
-                    return_schema["$ref"] = "#/definitions/" + schema.name
+                    # Check if we can use an enum
+                    if all(priority == COMPARABLE for priority in [_priority(value) for value in s.args]):
+                        or_values = [str(s) if isinstance(s, Literal) else s for s in s.args]
+                        # All values are simple, can use enum or const
+                        if len(or_values) == 1:
+                            return_schema["const"] = _to_json_type(or_values[0])
+                            return return_schema
+                        return_schema["enum"] = or_values
+                    else:
+                        # No enum, let's go with recursive calls
+                        any_of_values = []
+                        for or_key in s.args:
+                            new_value = _json_schema(_to_schema(or_key, i), is_main_schema=False)
+                            if new_value != {} and new_value not in any_of_values:
+                                any_of_values.append(new_value)
+                        if len(any_of_values) == 1:
+                            # Only one representable condition remains, do not put under anyOf
+                            return_schema.update(any_of_values[0])
+                        else:
+                            return_schema["anyOf"] = any_of_values
+                elif isinstance(s, And):
+                    # Handle And values
+                    all_of_values = []
+                    for and_key in s.args:
+                        new_value = _json_schema(_to_schema(and_key, i), is_main_schema=False)
+                        if new_value != {} and new_value not in all_of_values:
+                            all_of_values.append(new_value)
+                    if len(all_of_values) == 1:
+                        # Only one representable condition remains, do not put under allOf
+                        return_schema.update(all_of_values[0])
+                    else:
+                        return_schema["allOf"] = all_of_values
+                elif flavor == COMPARABLE:
+                    return_schema["const"] = _to_json_type(s)
+                elif flavor == VALIDATOR and type(s) == Regex:
+                    return_schema["type"] = "string"
+                    return_schema["pattern"] = s.pattern_str
                 else:
+                    if flavor != DICT:
+                        # If not handled, do not check
+                        return return_schema
+
+                    # Schema is a dict
+
                     required_keys = []
                     expanded_schema = {}
+                    additional_properties = i
                     for key in s:
                         if isinstance(key, Hook):
                             continue
 
+                        def _key_allows_additional_properties(key):
+                            """Check if a key is broad enough to allow additional properties"""
+                            if isinstance(key, Optional):
+                                return _key_allows_additional_properties(key.schema)
+
+                            return key == str or key == object
+
                         def _get_key_description(key):
                             """Get the description associated to a key (as specified in a Literal object). Return None if not a Literal"""
                             if isinstance(key, Optional):
@@ -636,6 +648,7 @@ class Schema(object):
 
                             return key
 
+                        additional_properties = additional_properties or _key_allows_additional_properties(key)
                         sub_schema = _to_schema(s[key], ignore_extra_keys=i)
                         key_name = _get_key_name(key)
 
@@ -646,7 +659,7 @@ class Schema(object):
                                 sub_schema, is_main_schema=False, description=_get_key_description(key)
                             )
                             if isinstance(key, Optional) and hasattr(key, "default"):
-                                expanded_schema[key_name]["default"] = _to_json_type(key.default)
+                                expanded_schema[key_name]["default"] = _to_json_type(_invoke_with_optional_kwargs(key.default, **kwargs) if callable(key.default) else key.default)
                         elif isinstance(key_name, Or):
                             # JSON schema does not support having a key named one name or another, so we just add both options
                             # This is less strict because we cannot enforce that one or the other is required
@@ -661,19 +674,19 @@ class Schema(object):
                             "type": "object",
                             "properties": expanded_schema,
                             "required": required_keys,
-                            "additionalProperties": i,
+                            "additionalProperties": additional_properties,
                         }
                     )
 
-                if is_main_schema:
-                    return_schema.update({"$id": schema_id, "$schema": "http://json-schema.org/draft-07/schema#"})
-                    if self._name:
-                        return_schema["title"] = self._name
+            if is_main_schema:
+                return_schema.update({"$id": schema_id, "$schema": "http://json-schema.org/draft-07/schema#"})
+                if self._name:
+                    return_schema["title"] = self._name
 
-                    if definitions_by_name:
-                        return_schema["definitions"] = {}
-                        for definition_name, definition in definitions_by_name.items():
-                            return_schema["definitions"][definition_name] = definition
+                if definitions_by_name:
+                    return_schema["definitions"] = {}
+                    for definition_name, definition in definitions_by_name.items():
+                        return_schema["definitions"][definition_name] = definition
 
             return _create_or_use_ref(return_schema)
 
@@ -752,8 +765,8 @@ class Literal(object):
 
 
 class Const(Schema):
-    def validate(self, data):
-        super(Const, self).validate(data)
+    def validate(self, data, **kwargs):
+        super(Const, self).validate(data, **kwargs)
         return data
 
 
